@@ -2,7 +2,7 @@ package Data::Reuse;
 
 # set up version info
 BEGIN {
-    $VERSION = '0.04';
+    $VERSION = '0.05';
 }    #BEGIN
 
 # be as strict and verbose as possible
@@ -12,8 +12,10 @@ use warnings;
 # we need this otherwise nothing works
 use Data::Alias qw(alias copy);
 
-# needed for creation of unique keys
+# other modules we need
+use Carp qw(croak);
 use Digest::MD5 qw(md5);
+use Scalar::Util qw(reftype);
 
 =for Explanation:
      Since Data::Alias uses Exporter, we might as well do that also.  Otherwise
@@ -22,8 +24,8 @@ use Digest::MD5 qw(md5);
 =cut
 
 use base 'Exporter';
-our @EXPORT      = qw(reuse);
-our @EXPORT_OK   = qw(reuse alias copy);
+our @EXPORT      = qw();
+our @EXPORT_OK   = qw(fixate forget reuse spread);
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
 
 # ID prefixes
@@ -33,10 +35,11 @@ my $A = "\1A\1";
 my $H = "\1H\1";
 
 # set up data store with predefined ro undef value
-my %reuse = ( $U => undef ); # must NOT use alias
+my %reuse;
+forget();
 
 # mark constants as read only
-Internals::SvREADONLY $_, 1 foreach ( $U, $S, $A, $H, $reuse{$U} );
+Internals::SvREADONLY $_, 1 foreach ( $U, $S, $A, $H );
 
 # recursion level
 my $level = 0;
@@ -50,12 +53,17 @@ my %handling;
 #---------------------------------------------------------------------------
 # reuse
 #
+# Make given values, store them in the constant hash and return them
+#
 #  IN: 1..N values to be folded into constants
 # OUT: 1..N same as input, but read-only and folded
 
 sub reuse (@); # needed because of recursion
 sub reuse (@) {
 
+    # being called with return values
+    my $not_void = defined wantarray;
+
     # we're one level deeper
     $level++;
 
@@ -72,7 +80,7 @@ sub reuse (@) {
                 }
 
                 # handle references
-                elsif ( my $ref = ref() ) {
+                elsif ( my $reftype = reftype $_ ) {
                     $handling{$_} = $level;
                     my $id;
 
@@ -80,13 +88,52 @@ sub reuse (@) {
                     alias {
 
                         # all elements of list
-                        if ( $ref eq 'ARRAY' ) {
+                        if ( $reftype eq 'ARRAY' ) {
                             $id = _list_id( $_ );
 
                             # not seen, recurse
                             if ( !exists $reuse{$id} ) {
                                 my @list = @{$_};
                                 (@list) = reuse @list;
+
+=for Explanation:
+     We need to use copy semantics, because aliasing semantics segfaults.
+     Or as Matthijs put it:
+=
+ Hah, het is een combinatie van:
+ 1. bug in perl m.b.t. \-prototypes (\[$%@] in dit geval)
+ 2. de refgen operator van Data::Alias (de impliciete \)
+ 3. het in void context zijn van dit blok
+ 4. totaal gebrek aan argument-checking in Internals::SvREADONLY
+ het prototype maakt van:
+=
+   Internals::SvREADONLY my @array, 1
+=
+ zoiets als:
+=
+   &Internals::SvREADONLY(\my @array, 1);
+=
+ Echter hij markeert de \ zonder enige context.  Dit hoort normaal alleen
+ de gebeuren voor de laatste expressie in een block, en betekent "evalueer
+ dit in de context van deze block".  In dit geval is het omliggende block
+ de top-level code, en die is altijd in void context.  Perl evalueert dus
+ de \ in void context.
+=
+ Hier ben je echter in perl zelf geen last van, omdat perl's ingebouwde
+ refgen op (\) alleen maar test op list-context, en gaat er van uit dat
+ het anders scalar context is.  D::A's refgen onderscheid alle drie de
+ contexts, en produceert dus niets omdat het in void context is.
+=
+ Hierdoor wordt de call dus:
+=
+   &Internals::SvREADONLY(1);
+=
+ En zoals ik al zei, Internals::SvREADONLY heeft geen argument validatie
+ en probeert dus fijn op adres 0x00000001 te lezen.. SEGV
+
+=cut
+
+                                # mark readonly, see above
                                 copy Internals::SvREADONLY @list, 1;
 
                                 # recursive structures may be replaced
@@ -95,13 +142,15 @@ sub reuse (@) {
                         }
 
                         # all values of hash
-                        elsif ( $ref eq 'HASH' ) {
+                        elsif ( $reftype eq 'HASH' ) {
                             $id = _hash_id( $_ );
 
                             # not seen, recurse, set result if first
                             if ( !$reuse{$id} ) {
                                 my %hash = %{$_};
                                 ( @hash{ keys %hash } ) = reuse values %hash;
+
+                                # mark readonly, see above
                                 copy Internals::SvREADONLY %hash, 1;
 
                                 # recursive structures may be replaced
@@ -110,7 +159,7 @@ sub reuse (@) {
                         }
 
                         # the value of a scalar ref
-                        elsif ( $ref eq 'SCALAR' ) {
+                        elsif ( $reftype eq 'SCALAR' ) {
                             my $scalar = ${$_};
 
                             # may be reused
@@ -135,7 +184,7 @@ sub reuse (@) {
 
                         # huh?
                         else {
-                            die "Cannot handle references of type '$ref'";
+                            croak "Cannot reuse '$reftype' references";
                         }
 
 =for Explanation:
@@ -147,7 +196,7 @@ sub reuse (@) {
 
 =cut
 
-                        $reuse{$id} ||= $_ if defined wantarray;
+                        $reuse{$id} ||= $_ if $not_void;
 
                         # store in data store
                         $reuse{$_} = $reuse{$id} || $_;
@@ -157,7 +206,11 @@ sub reuse (@) {
                     delete $handling{$_};
                 }
 
-                # not a ref
+                # not a ref, but now already in store
+                elsif ( exists $reuse{$_} ) {
+                }
+
+                # not a ref, and not in store either
                 else {
 
                     # not readonly already, make a read only copy
@@ -175,10 +228,130 @@ sub reuse (@) {
     $level--;
 
     # return aliases of the specified values if needed
-    alias return @reuse{ map { defined() ? $_ : $U } @_ }
-      if defined wantarray;
+    alias return @reuse{ map { defined() ? $_ : $U } @_ } if $not_void;
 }    #reuse
 
+#---------------------------------------------------------------------------
+# fixate
+#
+# Fixate the values of the given hash / array ref
+#
+#  IN: 1 hash / array ref
+#      2..N values for fixation
+
+sub fixate (\[@%]@) {
+
+    # fetch structure
+    alias my $struct = shift;
+    croak "Must specify a hash or array as first parameter to fixate"
+      unless my $reftype = reftype $struct;
+
+    # just fixate existing structure
+    reuse($struct), return if !@_;
+
+    # alias semantices from here on
+    alias {
+
+        # it's a hash
+        if ( $reftype eq 'HASH' ) {
+            my %hash = %{$struct};
+            croak "Can only fixate specific values on an empty hash"
+              if keys %hash;
+
+            # fill the hash and make sure only its values are reused
+            (%hash) = @_;
+            reuse \%hash; # also makes hash ro
+        }
+
+        # it's is an array
+        elsif ( $reftype eq 'ARRAY' ) {
+            my @array = @{$struct};
+            croak "Can only fixate specific values on an empty array"
+              if @array;
+
+            # fill the array and make sure its values are reused
+            (@array) = reuse @_;
+            copy Internals::SvREADONLY @array, 1; # must copy, see above
+        }
+
+        # huh?
+        else {
+            croak "Don't know how to fixate '$reftype' references";
+        }
+    };
+} #fixate
+
+#---------------------------------------------------------------------------
+# spread
+#
+# Spread a shared constant value in a data structure
+#
+#  IN: 1 data structure (hash / array ref)
+#      2 value to be set (default: undef )
+#      3..N keys / indexes to set
+
+sub spread (\[@%]@) {
+
+    # find out where to spread
+    alias my $struct = shift;
+    croak "Must specify a hash or array as first parameter to spread"
+      unless my $reftype = reftype $struct;
+
+    # huh? no value?
+    croak "Must specify a value as second parameter to spread"
+      if !@_;
+
+    # fetch proper constant alias
+    alias my $value = reuse shift;
+
+    # nothing to be done
+    return if !@_;
+
+    # alias semantics from here on
+    alias {
+
+        # it's a hash, but can we do it?
+        if ( $reftype eq 'HASH' ) {
+            my %hash = %{$struct};
+            croak "Cannot spread values in a restricted hash"
+              if Internals::SvREADONLY %hash;
+
+            # spread the values in the hash
+            $hash{$_} = $value foreach @_;
+        }
+
+        # it's an array, but can we do it?
+        elsif ( $reftype eq 'ARRAY' ) {
+            my @array = @{$struct};
+            croak "Cannot spread values in a restricted array"
+              if Internals::SvREADONLY @array;
+
+            # spread the values in the list
+            $array[$_] = $value foreach @_;
+        }
+
+        # huh?
+        else {
+            croak "Don't know how to spread values in '$reftype' references";
+        }
+    };
+}    #spread
+
+#---------------------------------------------------------------------------
+# forget
+#
+# Forget about the values that have been reused so far, or since the last
+# time "forget" was called.
+
+sub forget {
+
+    # copy a fresh undef value (shouldn't alias the system undef!)
+    %reuse = ( $U => undef );
+
+    # make sure this undef can't be changed
+    Internals::SvREADONLY $reuse{$U}, 1;
+} #forget
+
 #---------------------------------------------------------------------------
 #
 # Internal methods
@@ -217,13 +390,13 @@ sub _list_id {
 # Debug methods
 #
 #---------------------------------------------------------------------------
-# _reuse
+# _constants
 #
 # Return hash ref of hash containing the constant values
 #
 # OUT: 1 hash ref
 
-sub _reuse { return \%reuse } #_reuse
+sub _constants { return \%reuse } #_constants
 
 #---------------------------------------------------------------------------
 
@@ -233,28 +406,48 @@ __END__
 
 Data::Reuse - share constant values with Data::Alias
 
+=head1 VERSION
+
+This documentation describes version 0.05.
+
 =head1 SYNOPSIS
 
+ use Data::Reuse qw(fixate);
+ fixate my @array => ( 0, 1, 2, 3 );  # initialize and fixate
+ my @filled_array =  ( 0, 1, 2, 3 );
+ fixate @filled_array;                # fixate existing values
+ print \$array[0] == \$filled_array[0]
+   ? "Share memory\n" : "Don't share memory\n";
+
+ fixate my %hash => ( zero => 0, one => 1, two => 2, three => 3 ); 
+ my %filled_hash =  ( zero => 0, one => 1, two => 2, three => 3 );
+ fixate %filled_hash;
+ print \$hash{zero} == \$filled_hash{zero}
+   ? "Share memory\n" : "Don't share memory\n";
+
  use Data::Reuse qw(reuse);
  reuse my $listref = [ 0, 1, 2, 3 ];
  reuse my $hashref = { zero => 0, one => 1, two => 2, three => 3 };
  print \$listref->[0] == \$hashref->{zero}
    ? "Share memory\n" : "Don't share memory\n";
 
+ use Data::Alias qw(alias);
  use Data::Reuse qw(reuse);
- my @list = ( 0, 1, 2, 3 );
- my %hash = ( zero => 0, one => 1, two => 2, three => 3 );
- reuse \@list, \%hash;
- print \$list[0] == \$hash{zero}
-   ? "Share memory\n" : "Don't share memory\n";
-
- use Data::Reuse qw(reuse alias);  # use alias semantics from Data::Alias
  alias my @foo = reuse ( 0, 1, 2, 3 );
- alias my %bar = reuse ( zero => 0, one => 1, two => 2, three => 3 );
-
- print \$foo[0] == \$bar{zero}
+ print \$foo[0] == \$hashref->{zero}
    ? "Share memory\n" : "Don't share memory\n";
 
+ use Data::Reuse qw(spread);
+ spread my %spread_hash => undef, qw(foo bar baz);
+ print \$spread_hash{foo} == \$spread_hash{bar}
+   ? "Share memory\n" : "Don't share memory\n";
+ spread my @spread_array => 1, 0 .. 3;
+ print \$spread_array[0] == \$spread_array[1]
+   ? "Share memory\n" : "Don't share memory\n";
+
+ use Data::Reuse qw(forget);
+ forget();  # free memory used for tracking constant values
+
 =head1 DESCRIPTION
 
 By default, Perl doesn't share literal ( 0, 'foo' , "bar" ) values.  That's
@@ -277,13 +470,32 @@ allow you to have literal values share the same memory locations.
 Comes in this module, the L<Data::Reuse> module, which allows you to easily
 have literal and read-only values share the same memory address.  Which can
 save you a lot of memory when you are working with large data structures with
-similar values.  Which is especially nice in a mod_perl environment.
+similar values.  Which is especially nice in a mod_perl environment, where
+memory usage of persistent processes is one of the major issues..
 
 Of course, no memory savings will occur for literal values that only occur
-once.  So it is important that you use the "reuse"
+once.  So it is important that you use the functionality of this module
+wisely, only on values that you expect to be duplicated at least two times.
 
 =head1 SUBROUTINES
 
+=head2 fixate
+
+ fixate my @array => ( 0, 1, 2, 3 );
+
+ my @filled_array = ( 0, 1, 2, 3 );
+ fixate @filled_array;
+
+ fixate my %hash => ( zero => 0, one => 1, two => 2, three => 3 );
+
+ my %filled_hash = ( zero => 0, one => 1, two => 2, three => 3 );
+ fixate %filled_hash;
+
+The C<fixate> function allows you to initialize an array or hash with the given
+values, or to reuse all values in either an existing hash or an existing array,
+and making that hash or list read-only.  It is a frontend to C<reuse> and is
+mainly made for convenience only.
+
 =head2 reuse
 
  my $listref = reuse [ 1, 2, 3 ];
@@ -293,27 +505,28 @@ once.  So it is important that you use the "reuse"
  my %hash = ( one => 1, two > 2, three => 3 );
  reuse \@list, \%hash;
 
-The "reuse" function is the workhorse of this module.  Exported by default.
-It will investigate the given data structures and reuse any literal values
-as much as possible and return aliases to the given data structures.
+The C<reuse> function is the workhorse of this module.  It will investigate
+the given data structures and reuse any literal values as much as possible
+and mark the structure as read only and return aliases to the given data
+structures.
 
-=head2 alias
+=head2 spread
 
- use Data::Reuse qw(alias);
- alias my @list = reuse ( 1, 2, 3 );
- alias my %hash = reuse ( one => 1, two > 2, three => 3 );
+ spread @array => 1, ( 0, 1, 2, 3 );
 
-The "alias" semantics from the L<Data::Alias> module, not exported by default.
-Can be imported from L<Data::Reuse> for your convenience.  Please see the
-L<Data::Alias> module's documentation for more information.
+ spread %hash => undef, qw(foo bar baz);
 
-=head2 copy
+The C<spread> function allows you to quickly spread a single value to a number
+of elements in a list (specified by indexes), or to spread a single value to
+the values of a hash, specified by a number of keys.  It is a frontend to
+C<reuse> and is mainly made because you cannot use undef in C<alias> semantics
+as a value in a hash.  In other words:
 
- use Data::Reuse qw(copy);
+ alias @hash{ qw(foo bar baz) } = ();
 
-The "copy" semantics from the L<Data::Alias> module, not exported by default.
-Can be imported from L<Data::Reuse> for your convenience.  Please see the
-L<Data::Alias> module's documentation for more information.
+doesn't work, instead use:
+
+ spread %hash => undef, qw(foo bar baz);
 
 =head1 EXAMPLE
 
@@ -364,9 +577,16 @@ reused values.  This also goes for references, which are reused recursively.
 For scalar values, the value itself is used as a key.  For references, an
 MD5 hash is used as the key.
 
-All values are then aliased to the values in the hash (using <Data::Alias>'s
+All values are then aliased to the values in the hash (using L<Data::Alias>'s
 C<alias> feature) and returns as aliases when needed.
 
+The C<fixate> and C<spread> functions are basically frontends for the C<reuse>
+subroutine.
+
+The C<forget> function simply resets the internal hash used for storing
+constant values, freeing all memory associated with it that isn't referenced
+anywhere else (a.k.a. usually the memory used by the keys).
+
 =head1 CAVEATS
 
 =head2 reuse lists and hashes
@@ -383,7 +603,7 @@ is functionally equivalent with:
 
 so, this will cause the values B<1>, B<2> and B<3> to be in the internal reused
 values hash, but the assignment of C<@list> will use new copies, thus
-annihiliting any memory savings.
+annihilating any memory savings.
 
 Alternately:
 
@@ -415,7 +635,7 @@ occurs once in memory.  Suggestions / Patches to achieve this feature are
 B<very> welcome!
 
 If this proves to be impossible to do, then probably we need to use MD5 strings
-for all values to reduce memory requirements.
+for all values to reduce memory requirements (at the expense of more CPU usage).
 
 =head1 ACKNOWLEDGEMENTS
 
